/*(LGPL)
-----------------------------------------------------------------
	ZeeSpace - 2D + Height Field Rendering Engine
-----------------------------------------------------------------
 * Copyright (C) 2005 David Olofson
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "zeespace.h"


/*---------------------------------------------------------------
	Pseudo-random number generator with cross-fade
---------------------------------------------------------------*/

void zs_RndReset(ZS_State *state, float balance)
{
	state->rnd1 = state->rnd2 = 16576;
	if(balance < 0.0f)
		balance = 0.0f;
	else if(balance > 1.0f)
		balance = 1.0f;
	state->rbal = balance * 16384;
}

/* Returns a 32 bit pseudo random number; generator 1 */
static inline int rnd1_get(ZS_State *state)
{
	state->rnd1 *= 1566083941UL;
	++state->rnd1;
	state->rnd1 &= 0xffffffffUL;    /* NOP on 32 bit machines */
	return state->rnd1 * (state->rnd1 >> 16);
}


/* Returns a 32 bit pseudo random number; generator 2 */
static inline int rnd2_get(ZS_State *state)
{
	state->rnd2 = 16807 * state->rnd2;
	state->rnd2 %= 4294967294UL;
	--state->rnd2;
	return state->rnd2;
}


/*
 * Grab a "random" value that is a cross-fade between the
 * N'th value from rnd1_get() and rnd2_get().
 */
static inline int rnd_get(ZS_State *state, unsigned range)
{
	unsigned div = 0xffffffff / range;
	unsigned r1 = rnd1_get(state) / div;
	unsigned r2 = rnd2_get(state) / div;
	return (r1 * state->rbal + r2 * (16383 - state->rbal)) >> 14;
}


int zs_RndGet(ZS_State *state, unsigned range)
{
	return rnd_get(state, range);
}


/*---------------------------------------------------------------
	Integer math
---------------------------------------------------------------*/

/* Round 'value' upwards by 'granularity' */
static inline int granup(int value, int granularity)
{
	return (value + granularity - 1) / granularity * granularity;
}

/* Clamp 'val' to the range of unsigned 8:8 fixed point */
static inline unsigned clamp88(int value)
{
	if(value < 0)
		return 0;
	else if(value > 65535)
		return 65535;
	else
		return value;
}


/* Square root for 24:8 fixed point */
static inline unsigned long sqrt248(unsigned long val)
{
	return (unsigned long)(sqrtf(val) + 128) >> 8;
}


/*---------------------------------------------------------------
	State management
---------------------------------------------------------------*/

/* Prepare internal buffers for a maximum span width of 'w' pixels */
static int zs_StatePrepare(ZS_State *st, int w)
{
	if(w <= st->bufw)
		return 0;
	free(st->scratch);
	free(st->pipebuf);
	st->scratch = (ZS_Pixel *)malloc(sizeof(ZS_Pixel) * w);
	st->pipebuf = (ZS_Pixel *)malloc(sizeof(ZS_Pixel) * w);
	if(!st->scratch || !st->pipebuf)
		return -1;
	st->bufw = w;
	return 0;
}


ZS_State *zs_Open(void)
{
	ZS_State *st = (ZS_State *)calloc(1, sizeof(ZS_State));
	if(!st)
		return NULL;
	st->refcount = 1;
	if(zs_StatePrepare(st, 100) < 0)
	{
		free(st);
		return NULL;
	}
	return st;
}


void zs_Close(ZS_State *state)
{
	if(--state->refcount <= 0)
	{
		free(state->scratch);
		free(state->pipebuf);
		free(state);
	}
}


/*---------------------------------------------------------------
	Surface/Window/Region management
---------------------------------------------------------------*/

ZS_Surface *zs_NewSurface(ZS_State *state, int w, int h)
{
	ZS_Surface *s = calloc(1, sizeof(ZS_Surface));
	if(!s)
		return NULL;
	if(zs_StatePrepare(state, w) < 0)
	{
		free(s);
		return NULL;
	}
	s->refcount = 1;
	s->state = state;
	s->pixels = calloc(w * h, sizeof(ZS_Pixel));
	if(!s->pixels)
	{
		free(s);
		return NULL;
	}
	s->w = s->pitch = w;
	s->h = h;
	s->flags = ZS_SF_OWNSPIXELS;
	++state->refcount;
	return s;
}


ZS_Surface *zs_SubSurface(ZS_Surface *from, ZS_Rect *area)
{
	ZS_Surface *s = calloc(1, sizeof(ZS_Surface));
	if(!s)
		return NULL;
	s->state = from->state;
	s->parent = from;
	s->pitch = from->pitch;
	s->w = area->w;
	s->h = area->h;
	s->pixels = zs_Pixel(from, area->x, area->y);
	s->flags = 0;
	++from->refcount;
	++s->state->refcount;
	return s;
}


void zs_FreeSurface(ZS_Surface *s)
{
	if(--s->refcount > 0)
		return;
	if(s->flags & ZS_SF_OWNSPIXELS)
		free(s->pixels);
	else
		zs_FreeSurface(s->parent);
	zs_Close(s->state);
	free(s);
}


static inline int zs_Clip(ZS_Rect *r, int w, int h)
{
	int nw, nh;
	int x2 = r->x + r->w;
	int y2 = r->y + r->h;
	if(r->x < 0)
		r->x = 0;
	if(x2 > w)
		x2 = w;
	if(r->y < 0)
		r->y = 0;
	if(y2 > h)
		y2 = h;
	r->w = nw = x2 - r->x;
	if(r->w <= 0)
		return 0;
	r->h = nh = y2 - r->y;
	if(r->h <= 0)
		return 0;
	return 1;
}


void zs_RawFill(ZS_Channels channels, ZS_Surface *s, ZS_Rect *r, ZS_Pixel *p)
{
	ZS_Rect cr;
	int xmin, ymin, x, y, xmax, ymax;
	if(!r)
	{
		r = &cr;
		cr.x = cr.y = 0;
		cr.w = s->w;
		cr.h = s->h;
	}
	else if(!zs_Clip(r, s->w, s->h))
		return;
	xmin = r->x;
	ymin = r->y;
	xmax = r->x + r->w;
	ymax = r->y + r->h;
	channels &= ZS_CH_ALL;
	for(y = ymin; y < ymax; ++y)
	{
		ZS_Pixel *lp = zs_Pixel(s, 0, y);
		switch(channels)
		{
		  case ZS_CH_ALL:
			for(x = xmin; x < xmax; ++x)
				lp[x] = *p;
			break;
		  case ZS_CH_RGB:
			for(x = xmin; x < xmax; ++x)
			{
				lp[x].r = p->r;
				lp[x].g = p->g;
				lp[x].b = p->b;
			}
			break;
		  case ZS_CH_RGBA:
			for(x = xmin; x < xmax; ++x)
			{
				lp[x].r = p->r;
				lp[x].g = p->g;
				lp[x].b = p->b;
				lp[x].a = p->a;
			}
			break;
		  case ZS_CH_RGB | ZS_CH_Z:
			for(x = xmin; x < xmax; ++x)
			{
				lp[x].r = p->r;
				lp[x].g = p->g;
				lp[x].b = p->b;
				lp[x].z = p->z;
			}
			break;
		  default:
			/* Slow(er) catch-all */
			if(channels & ZS_CH_R)
				for(x = xmin; x < xmax; ++x)
					lp[x].r = p->r;
			if(channels & ZS_CH_G)
				for(x = xmin; x < xmax; ++x)
					lp[x].g = p->g;
			if(channels & ZS_CH_B)
				for(x = xmin; x < xmax; ++x)
					lp[x].b = p->b;
			if(channels & ZS_CH_A)
				for(x = xmin; x < xmax; ++x)
					lp[x].a = p->a;
			if(channels & ZS_CH_I)
				for(x = xmin; x < xmax; ++x)
					lp[x].i = p->i;
			if(channels & ZS_CH_Z)
				for(x = xmin; x < xmax; ++x)
					lp[x].z = p->z;
			break;
		}
	}
}


void zs_RawCopy(ZS_Channels channels, ZS_Surface *src, ZS_Rect *srcr,
		ZS_Surface *dst, int dx, int dy)
{
	ZS_Rect cr;
	int x, y;

	/* Source rect */
	if(srcr)
	{
		/* Clipping */
		if(!zs_Clip(srcr, src->w, src->h))
			return;
	}
	else
	{
		/* Use full surface (no clipping needed) */
		srcr = &cr;
		cr.x = cr.y = 0;
		cr.w = src->w;
		cr.h = src->h;
	}

	/* Destination clipping */
/* TODO: */

	channels &= ZS_CH_ALL;
	for(y = 0; y < srcr->h; ++y)
	{
		ZS_Pixel *sp = zs_Pixel(src, srcr->x, srcr->y + y);
		ZS_Pixel *dp = zs_Pixel(dst, dx, dy + y);
		switch(channels)
		{
		  case ZS_CH_ALL:
		  {
			int bytes = sizeof(ZS_Pixel) * srcr->w;
			memcpy(dp, sp, bytes);
			break;
		  }
		  default:
			/* Slow(er) catch-all */
			if(channels & ZS_CH_R)
				for(x = 0; x < srcr->w; ++x)
					dp[x].r = sp[x].r;
			if(channels & ZS_CH_G)
				for(x = 0; x < srcr->w; ++x)
					dp[x].g = sp[x].g;
			if(channels & ZS_CH_B)
				for(x = 0; x < srcr->w; ++x)
					dp[x].b = sp[x].b;
			if(channels & ZS_CH_A)
				for(x = 0; x < srcr->w; ++x)
					dp[x].a = sp[x].a;
			if(channels & ZS_CH_I)
				for(x = 0; x < srcr->w; ++x)
					dp[x].i = sp[x].i;
			if(channels & ZS_CH_Z)
				for(x = 0; x < srcr->w; ++x)
					dp[x].z = sp[x].z;
			break;
		}
	}
}


/*
 * Simulate air/water transition by compressing Z below the surface.
 * Apply *before* lighting!
 */
void zs_WaterZ(ZS_Surface *s, ZS_Rect *r, ZS_Pixel *p, int factor)
{
	ZS_Rect cr;
	int xmin, ymin, x, y, xmax, ymax;
	if(!r)
	{
		r = &cr;
		cr.x = cr.y = 0;
		cr.w = s->w;
		cr.h = s->h;
	}
	else if(!zs_Clip(r, s->w, s->h))
		return;
	xmin = r->x;
	ymin = r->y;
	xmax = r->x + r->w;
	ymax = r->y + r->h;
	for(y = ymin; y < ymax; ++y)
	{
		ZS_Pixel *lp = zs_Pixel(s, 0, y);
		for(x = xmin; x < xmax; ++x)
			if(p->z >= lp[x].z)
				lp[x].z += (p->z - lp[x].z) * factor >> 8;
	}
}


void zs_Fog(ZS_Surface *s, ZS_Rect *r, ZS_Pixel *p, int visibility)
{
	ZS_Rect cr;
	int xmin, ymin, x, y, xmax, ymax;
	int blur = visibility > 0 ? 65536 / visibility : 65536;
	if(!r)
	{
		r = &cr;
		cr.x = cr.y = 0;
		cr.w = s->w;
		cr.h = s->h;
	}
	else if(!zs_Clip(r, s->w, s->h))
		return;
	xmin = r->x;
	ymin = r->y;
	xmax = r->x + r->w;
	ymax = r->y + r->h;
	for(y = ymin; y < ymax; ++y)
	{
		ZS_Pixel *lp = zs_Pixel(s, 0, y);
		for(x = xmin; x < xmax; ++x)
			if(p->z >= lp[x].z)
			{
				int b1, b2;
				b1 = p->a + ((int)(p->z - lp[x].z) * blur >> 16);
				if(b1 > 256)
					b1 = 256;
				b2 = 256 - b1;
				lp[x].r = (p->r * b1 + lp[x].r * b2) >> 8;
				lp[x].g = (p->g * b1 + lp[x].g * b2) >> 8;
				lp[x].b = (p->b * b1 + lp[x].b * b2) >> 8;
				lp[x].i = (p->i * b1 + lp[x].i * b2) >> 8;
			}
	}
}


void zs_Block(ZS_Pipe *pipe, ZS_Surface *s,
		float x, float y, float w, float h, ZS_Pixel *px)
{
	ZS_Pixel *scratch = s->state->scratch;
	int ix, iy;
	int th = px->z;
	int ymin = y;
	int ymax = y + h;
	int xmin = x;
	int xmax = x + w;
	if(xmin < 0)
		xmin = 0;
	if(xmax > s->w)
		xmax = s->w;
	if(ymin < 0)
		ymin = 0;
	if(ymax > s->h)
		ymax = s->h;
	for(iy = ymin; iy < ymax; ++iy)
	{
		int w = xmax - xmin;
		for(ix = 0; ix < w; ++ix)
		{
			scratch[ix].r = px->r;
			scratch[ix].g = px->g;
			scratch[ix].b = px->b;
			scratch[ix].z = th;
		}
		zs_PipeRun(pipe, 0, scratch, zs_Pixel(s, xmin, iy), w);
	}
}


void zs_Cylinder(ZS_Pipe *pipe, ZS_Surface *s,
		float cx, float cy, float r, ZS_Pixel *px)
{
	ZS_Pixel *scratch = s->state->scratch;
	int x, y;
	int ymin = cy - r;
	int ymax = cy + r;
	if(ymin < 0)
		ymin = 0;
	if(ymax > s->h)
		ymax = s->h;
	for(y = ymin; y < ymax; ++y)
	{
		float dydy = (y - cy)*(y - cy);
		float hw = sqrtf(r*r - dydy);
		int xmin = cx - hw;
		int xmax = cx + hw;
		int w;
		if(xmin < 0)
			xmin = 0;
		if(xmax > s->w)
			xmax = s->w;
		w = xmax - xmin;
		for(x = 0; x < w; ++x)
		{
			scratch[x].r = px->r;
			scratch[x].g = px->g;
			scratch[x].b = px->b;
			scratch[x].z = px->z;
		}
		zs_PipeRun(pipe, 0, scratch, zs_Pixel(s, xmin, y), w);
	}
}


void zs_Cone(ZS_Pipe *pipe, ZS_Surface *s,
		float cx, float cy, float r, float h, ZS_Pixel *px)
{
	ZS_Pixel *scratch = s->state->scratch;
	float th = px->z + h;
	int x, y;
	float fx, fy;
	int ymin = cy - r;
	int ymax = cy + r;
	if(ymin < 0)
		ymin = 0;
	if(ymax > s->h)
		ymax = s->h;
	for(fy = y = ymin; y < ymax; ++y, ++fy)
	{
		float dydy = (fy - cy)*(fy - cy);
		float hw = sqrtf(r*r - dydy);
		int xmin = cx - hw;
		int xmax = cx + hw;
		int w;
		if(xmin < 0)
			xmin = 0;
		if(xmax > s->w)
			xmax = s->w;
		w = xmax - xmin;
		for(x = 0, fx = xmin; x < w; ++x, ++fx)
		{
			float dx = fx - cx;
			float d = sqrtf(dx*dx + dydy);
			int z = th - (h * d) / r;
			scratch[x].r = px->r;
			scratch[x].g = px->g;
			scratch[x].b = px->b;
			scratch[x].z = clamp88(z);
		}
		zs_PipeRun(pipe, 0, scratch, zs_Pixel(s, xmin, y), w);
	}
}


void zs_Beehive(ZS_Pipe *pipe, ZS_Surface *s,
		float cx, float cy, float r, float h, ZS_Pixel *px)
{
	ZS_Pixel *scratch = s->state->scratch;
	int x, y;
	float fx, fy;
	int ymin = cy - r;
	int ymax = cy + r;
	if(ymin < 0)
		ymin = 0;
	if(ymax > s->h)
		ymax = s->h;
	for(fy = y = ymin; y < ymax; ++y, ++fy)
	{
		float dydy = (fy - cy)*(fy - cy);
		float hw = sqrtf(r*r - dydy);
		int xmin = cx - hw;
		int xmax = cx + hw;
		int w;
		if(xmin < 0)
			xmin = 0;
		if(xmax > s->w)
			xmax = s->w;
		w = xmax - xmin;
		for(x = 0, fx = xmin; x < w; ++x, ++fx)
		{
			float dx = fx - cx;
			float d = sqrtf(dx*dx + dydy);
			int z = ((r - d * d / r) * h / r) * 256.0 + px->z;
			scratch[x].r = px->r;
			scratch[x].g = px->g;
			scratch[x].b = px->b;
			scratch[x].z = clamp88(z);
		}
		zs_PipeRun(pipe, 0, scratch, zs_Pixel(s, xmin, y), w);
	}
}


void zs_Dome(ZS_Pipe *pipe, ZS_Surface *s,
		float cx, float cy, float r, float h, ZS_Pixel *px)
{
	ZS_Pixel *scratch = s->state->scratch;
	int x, y;
	float fx, fy;
	float rr = r*r;
	int ymin = cy - r;
	int ymax = cy + r;
	float hf = h * 256.0 / r;
	float hz = px->z;
	if(ymin < 0)
		ymin = 0;
	if(ymax > s->h)
		ymax = s->h;
	for(fy = y = ymin; y < ymax; ++y, ++fy)
	{
		float dydy = (fy - cy)*(fy - cy);
		float hw = sqrtf(rr - dydy);
		int xmin = cx - hw;
		int xmax = cx + hw;
		int w;
		if(xmin < 0)
			xmin = 0;
		if(xmax > s->w)
			xmax = s->w;
		w = xmax - xmin;
		for(x = 0, fx = xmin; x < w; ++x, ++fx)
		{
			float dx = fx - cx;
			float d = sqrtf(dx*dx + dydy);
			int z = sqrtf(rr - d*d) * hf + hz;
			scratch[x].r = px->r;
			scratch[x].g = px->g;
			scratch[x].b = px->b;
			scratch[x].z = clamp88(z);
		}
		zs_PipeRun(pipe, 0, scratch, zs_Pixel(s, xmin, y), w);
	}
}


/*---------------------------------------------------------------
	ZeeSpace toolkit - SDL rendering
---------------------------------------------------------------*/
void zs_Blit2SDL(ZS_Surface *from, SDL_Surface *to)
{
	int x, y, xmax, ymax;
	int bpp = to->format->BytesPerPixel;
#ifdef SDL_LITTLE_ENDIAN
	int ro = 3 - to->format->Rshift / 8;
	int go = 3 - to->format->Gshift / 8;
	int bo = 3 - to->format->Bshift / 8;
#else
	int ro = to->format->Rshift / 8;
	int go = to->format->Gshift / 8;
	int bo = to->format->Bshift / 8;
#endif
	xmax = to->w;
	ymax = to->h;
	SDL_LockSurface(to);
	for(y = 0; y < ymax; ++y)
	{
		ZS_Pixel *p = zs_Pixel(from, 0, y);
		unsigned char *d = ((unsigned char *)to->pixels) + y * to->pitch;
		for(x = 0; x < xmax; ++x)
		{
			int r, g, b;
			r = p->r * p->i >> 8;
			g = p->g * p->i >> 8;
			b = p->b * p->i >> 8;
			if(r > 255)
				r = 255;
			if(g > 255)
				g = 255;
			if(b > 255)
				b = 255;
			d[ro] = r;
			d[go] = g;
			d[bo] = b;
			d += bpp;
			++p;
		}
	}
	SDL_UnlockSurface(to);
}


static inline void vline(SDL_Surface *s, ZS_Pixel *p, int x, int y, int y2,
		int ro, int go, int bo)
{
	unsigned char *d;
	int bpp = s->format->BytesPerPixel;
	int r, g, b, bal;
	r = p->r * p->i >> 8;
	g = p->g * p->i >> 8;
	b = p->b * p->i >> 8;
	if(r > 255)
		r = 255;
	if(g > 255)
		g = 255;
	if(b > 255)
		b = 255;
	y >>= 8;
	bal = y2 & 0xff;
	y2 >>= 8;
	d = ((unsigned char *)s->pixels) + y * s->pitch + x * bpp;
	while(++y <= y2)
	{
		d[ro] = r;
		d[go] = g;
		d[bo] = b;
		d += s->pitch;
	}
	if(y < s->h)
	{
		d[ro] = (r * bal + d[ro] * (255 - bal)) >> 8;
		d[go] = (g * bal + d[go] * (255 - bal)) >> 8;
		d[bo] = (b * bal + d[bo] * (255 - bal)) >> 8;
	}
}

/* NOTE: Wrecks the bottom pixel row of 'from'! */
void zs_3D2SDL(ZS_Surface *from, SDL_Surface *to, float pitch)
{
	ZS_Pixel sky;
	ZS_Pixel *p;
	float a = pitch * 2 * M_PI / 360;
	unsigned tilt, scale, ty;
#ifdef SDL_LITTLE_ENDIAN
	int ro = 3 - to->format->Rshift / 8;
	int go = 3 - to->format->Gshift / 8;
	int bo = 3 - to->format->Bshift / 8;
#else
	int ro = to->format->Rshift / 8;
	int go = to->format->Gshift / 8;
	int bo = to->format->Bshift / 8;
#endif
	int *buf = (int *)zs_Pixel(from, 0, from->h - 1);
	int x, y, xmax, ymax, ymaxf;
	xmax = to->w - 1;
	ymax = to->h - 1;
	ymaxf = ymax << 8;
	for(x = 0; x <= xmax; ++x)
		buf[x] = ymaxf + 256;
	SDL_LockSurface(to);
	y = ymax;
	if((ZS_Pixel *)buf == zs_Pixel(from, 0, y))
		--y;
	p = zs_Pixel(from, 0, y);
	for(x = 0; x <= xmax; ++x)
		p[x].i = 0;
	tilt = sin(a) * 65536;
	scale = cos(a) * 65536;
	ty = 0;
	for( ; y >= 0; --y)
	{
		p = zs_Pixel(from, 0, y);
		for(x = 0; x <= xmax; ++x)
		{
			int sz = p[x].z * scale >> 8;
			int tz = ymaxf - ((ty  + sz) >> 8);
			if(tz < buf[x])
			{
				if(tz < 0)
					tz = 0;
				vline(to, &p[x], x, tz, buf[x], ro, go, bo);
				buf[x] = tz;
			}
		}
		ty += tilt;
	}
	sky.r = 64;
	sky.g = 64;
	sky.b = 96;
	sky.a = 255;
	sky.i = 256;
	for(x = 0; x <= xmax; ++x)
		vline(to, &sky, x, 0, buf[x], ro, go, bo);
	SDL_UnlockSurface(to);
}


void zs_Graph2SDL(ZS_Surface *from, SDL_Surface *to, int scale)
{
	int x, y, xmax, ymax;
	int bpp = to->format->BytesPerPixel;
#ifdef SDL_LITTLE_ENDIAN
	int ro = 3 - to->format->Rshift / 8;
	int go = 3 - to->format->Gshift / 8;
	int bo = 3 - to->format->Bshift / 8;
#else
	int ro = to->format->Rshift / 8;
	int go = to->format->Gshift / 8;
	int bo = to->format->Bshift / 8;
#endif
	xmax = to->w;
	ymax = to->h;
	SDL_LockSurface(to);
	for(y = 0; y < ymax; ++y)
	{
		ZS_Pixel *p = zs_Pixel(from, 0, y / scale * scale);
		unsigned char *d = ((unsigned char *)to->pixels) + y * to->pitch;
		for(x = 0; x < xmax; ++x)
		{
			unsigned g, fy, sfy;
			fy = (65535 - p->z) * scale >> 8;
			sfy = fy >> 8;
			if(y % scale > sfy)
				g = 255;
			else if(y % scale < sfy)
				g = 0;
			else
				g = ((y % scale << 8) - fy) & 0xff;
			d[ro] = y % scale == scale - 1 ? 255 : 0;
			d[go] = g;
			d[bo] = y % scale == scale / 2 ? 255 : 0;
			d += bpp;
			++p;
		}
	}
	SDL_UnlockSurface(to);
}


void zs_Z2SDL(ZS_Surface *from, SDL_Surface *to, int scale)
{
	int x, y, xmax, ymax;
	int bpp = to->format->BytesPerPixel;
#ifdef SDL_LITTLE_ENDIAN
	int ro = 3 - to->format->Rshift / 8;
	int go = 3 - to->format->Gshift / 8;
	int bo = 3 - to->format->Bshift / 8;
#else
	int ro = to->format->Rshift / 8;
	int go = to->format->Gshift / 8;
	int bo = to->format->Bshift / 8;
#endif
	xmax = to->w;
	ymax = to->h;	
	SDL_LockSurface(to);
	for(y = 0; y < ymax; ++y)
	{
		ZS_Pixel *p = zs_Pixel(from, 0, y);
		unsigned char *d = ((unsigned char *)to->pixels) + y * to->pitch;
		for(x = 0; x < xmax; ++x, d += bpp, ++p)
		{
			int c;
			int rawc = p->z;
			rawc *= scale;
			rawc >>= 6 + 6;
			if(rawc > 1023)
			{
				d[ro] = 255;
				d[go] = 0;
				d[bo] = 255;
				continue;
			}
			c = rawc & 0xff;
			switch(rawc >> 8)
			{
			  case 0:
				d[ro] = 0;
				d[go] = 0;
				d[bo] = c;
				break;
			  case 1:
				d[ro] = 0;
				d[go] = c;
				d[bo] = 255 - c;
				break;
			  case 2:
				d[ro] = c;
				d[go] = 255;
				d[bo] = 0;
				break;
			  case 3:
				d[ro] = 255;
				d[go] = 255;
				d[bo] = c;
				break;
			}
		}
	}
	SDL_UnlockSurface(to);
}


void zs_I2SDL(ZS_Surface *from, SDL_Surface *to)
{
	int x, y, xmax, ymax;
	int bpp = to->format->BytesPerPixel;
#ifdef SDL_LITTLE_ENDIAN
	int ro = 3 - to->format->Rshift / 8;
	int go = 3 - to->format->Gshift / 8;
	int bo = 3 - to->format->Bshift / 8;
#else
	int ro = to->format->Rshift / 8;
	int go = to->format->Gshift / 8;
	int bo = to->format->Bshift / 8;
#endif
	xmax = to->w;
	ymax = to->h;	
	SDL_LockSurface(to);
	for(y = 0; y < ymax; ++y)
	{
		ZS_Pixel *p = zs_Pixel(from, 0, y);
		unsigned char *d = ((unsigned char *)to->pixels) + y * to->pitch;
		for(x = 0; x < xmax; ++x, d += bpp, ++p)
		{
			int i = p->i;
			if(i <= 255)
			{
				d[ro] = 0;
				d[go] = i;
				d[bo] = 0;
			}
			else if(i <= 511)
			{
				d[ro] = i;
				d[go] = 255;
				d[bo] = 0;
			}
			else
				d[ro] = d[go] = d[bo] = 255;
		}
	}
	SDL_UnlockSurface(to);
}


/*---------------------------------------------------------------
	Shadow casting
---------------------------------------------------------------*/

/* Do one diagonal ray */
static inline void do_ray(ZS_Surface *s, int startx, int starty, int zstep)
{
	ZS_Pixel *p = zs_Pixel(s, startx, starty);
	int step = s->pitch + 1;
	int z = p->z;
	int i = s->w - startx;
	int ymax = s->h - starty;
	int diffusion = 8192;
	int hardness = 128;
 	if(i > ymax)
		i = ymax;
	while(i--)
	{
		int pz = p->z;
		z += zstep;
		if(z <= pz)
			z = pz;
		else
		{
			unsigned sh = z - pz;
			unsigned pd = p->i * diffusion / (diffusion + sh);
			unsigned ph = p->i * hardness / (hardness + sh);
			p->i = (ph + pd) >> 1;
		}
		p += step;
	}
}

/* Cast shadows by attenuating the I channel */
void zs_ShadowCast(ZS_Surface *s, float altitude, float ambient)
{
	int xmin, ymin, x, y, xmax, ymax, zstep;
	if(altitude >= 90)
		return;
 	zstep = -tan(altitude * 2 * M_PI / 360) * 256;
	xmin = 0;
	ymin = 0;
	xmax = s->w;
	ymax = s->h;
	for(y = ymin; y < ymax; ++y)
		do_ray(s, xmin, y, zstep);
	for(x = ymin + 1; x < xmax; ++x)
		do_ray(s, x, ymin, zstep);
}


/*---------------------------------------------------------------
	Bump Mapping
---------------------------------------------------------------*/
void zs_BumpMap(ZS_Surface *s, float altitude, float ambient)
{
	int mini = ambient * 256.0;
	int x, y, xmax, ymax, abias;
	if(altitude < 1)
		altitude = 1;
	abias = tan((90 - altitude) * 2 * M_PI / 360) * 256;
	xmax = s->w - 2;
	ymax = s->h - 2;
	for(y = 0; y <= ymax; ++y)
	{
		ZS_Pixel *p = zs_Pixel(s, 0, y);
		int pitch = s->pitch;
		int z00 = p[0].z;
		int z10 = p[pitch].z;
		for(x = 0; x <= xmax; ++x)
		{
			float face, face2;
			int iface;
			int z01 = p[x + 1].z;
			int z11 = p[x + pitch + 1].z;
			face2 = fabs(z01 - z10);
			face = fabs(z11 - z00 - abias);
			iface = (int)(sqrtf(face*face + face2*face2) * .5f);
			iface = 256 + 128 - iface;
			if(iface < mini)
				iface = mini;
			p[x].i = p[x].i * iface >> 8;
			z00 = z01;
			z10 = z11;
		}
	}
}


/*---------------------------------------------------------------
	Terrain Generator
---------------------------------------------------------------*/
static inline void addnoise(ZS_State *st, int base, int amp, ZS_Pixel *p)
{
	int z = base;
	if(!amp)
		rnd_get(st, 10);	/* Dummy. */
	else
		z += rnd_get(st, amp << 1) - amp;
	if(z < 0)
		p->z = 0;
	else if(z > 65535)
		p->z = 65535;
	else
		p->z = z;
}

static inline void seedline(ZS_Surface *s, int d, int ibase, int iamp, int y)
{
	int x;
	ZS_Pixel *p = zs_Pixel(s, 0, y);
	ZS_State *st = s->state;
	for(x = 0; x <= s->w - 1; x += d)
		addnoise(st, ibase, iamp, p + x);
	addnoise(st, ibase, iamp, p + s->w - 1);
}

static inline int filter_bilin(int x, int y, int a, int b)
{
	return (x + y + a + b) >> 2;
}

void zs_TerrainZ(ZS_Surface *s, float base, float amp, float acoeff,
		int tilesize)
{
	ZS_State *state = s->state;
	int x, y, d, xmax, ymax;
	int ibase = (int)(base * 256);
	int iamp = (int)(amp * 256);
	xmax = s->w - 1;
	ymax = s->h - 1;
	d = tilesize;

	/* Seed - one point in the top-left corner of each tile */
	for(y = 0; y <= s->h - 1; y += d)
		seedline(s, d, ibase, iamp, y);
	seedline(s, d, ibase, iamp, s->h - 1);
	
	d >>= 1;
	for( ; d; d >>= 1)
	{
		ZS_Pixel *p;
		amp = amp * acoeff;
		iamp = (int)(amp * 256);

		/*
		 * Top and bottom edges:
		 *	tl(st)tr ..
		 */
		{
			int tr = zs_PixelClamp(s, 0, 0)->z;
			p = zs_Pixel(s, d, 0);
			for(x = d; x <= xmax; x += 2 * d, p += 2*d)
			{
				int st;
				int tl = (tr);
				tr = zs_PixelClamp(s, x + d, 0)->z;
				st = filter_bilin(tl, tr, tr, tl);
				addnoise(state, st, iamp, p);
			}
			tr = zs_PixelClamp(s, 0, ymax)->z;
			p = zs_Pixel(s, d, ymax);
			for(x = d; x <= xmax; x += 2 * d, p += 2*d)
			{
				int st;
				int tl = (tr);
				tr = zs_PixelClamp(s, x + d, ymax)->z;
				st = filter_bilin(tl, tr, tr, tl);
				addnoise(state, st, iamp, p);
			}
		}

		/*
		 * Right edge:
		 *	 tr
		 *	(sr)
		 *	 br
		 *	 ..
		 */
		{
			int br = zs_PixelClamp(s, xmax, 0)->z;
			p = zs_Pixel(s, xmax, d);
			for(y = d; y <= ymax; y += 2 * d, p += 2 * d * s->pitch)
			{
				int sr;
				int tr = (br);
				br = zs_PixelClamp(s, xmax, y + d)->z;
				sr = filter_bilin(tr, br, br, tr);
				addnoise(state, sr, iamp, p);
			}
		}

		/*
		 * Step 1:
		 *	tl .. tr ..
		 *	..(s1).. ..
		 *	bl .. br ..
		 *	.. .. .. ..
		 */
		for(y = d; y <= ymax; y += 2 * d)
		{
			int tr = zs_PixelClamp(s, 0, y - d)->z;
			int br = zs_PixelClamp(s, 0, y + d)->z;
			p = zs_Pixel(s, d, y);
			for(x = d; x <= xmax; x += 2 * d, p += 2*d)
			{
				int s1;
				int tl = (tr);
				int bl = (br);
				tr = zs_PixelClamp(s, x + d, y - d)->z;
				br = zs_PixelClamp(s, x + d, y + d)->z;
				s1 = filter_bilin(tl, br, tr, bl);
				addnoise(state, s1, iamp, p);
			}
		}
		/*
		 * Step 2:
		 *	.. .. tr ..
		 *	.. s1(s2)rr
		 *	.. .. br ..
		 *	.. .. .. ..
		 */
		for(y = d; y <= ymax; y += 2 * d)
		{
			int rr = zs_PixelClamp(s, d, y)->z;
			p = zs_Pixel(s, 0, y);
			for(x = -d; x <= xmax - d; x += 2 * d, p += 2 * d)
			{
				int s2, br;
				int tr = zs_PixelClamp(s, x + d, y - d)->z;
				int s1 = (rr);
				rr = zs_PixelClamp(s, x + 2*d, y)->z;
				br = zs_PixelClamp(s, x + d, y + d)->z;
				s2 = filter_bilin(tr, br, s1, rr);
				addnoise(state, s2, iamp, p);
			}
		}
		/*
		 * Step 3:
		 *	.. .. .. ..
		 *	.. s1 .. ..
		 *	bl(s3)br ..
		 *	.. bb .. ..
		 */
		for(y = -d; y <= ymax - d; y += 2 * d)
		{
			int br = zs_PixelClamp(s, 0, y + d)->z;
			p = zs_Pixel(s, d, y + d);
			for(x = d; x <= xmax; x += 2 * d, p += 2 * d)
			{
				int s1, bl, s3, bb;
				s1 = zs_PixelClamp(s, x, y)->z;
				bl = (br);
				br = zs_PixelClamp(s, x + d, y + d)->z;
				bb = zs_PixelClamp(s, x, y + 2*d)->z;
				s3 = filter_bilin(s1, bb, bl, br);
				addnoise(state, s3, iamp, p);
			}
		}
	}
}
