/*(LGPL)
-----------------------------------------------------------------
	zs_pipe.c - ZeeSpace Pixel Processing Pipeline
-----------------------------------------------------------------
 * Copyright (C) 2005 David Olofson
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "zeespace.h"

/*------------------------------------------------------------------
	Pipe implementation
------------------------------------------------------------------*/

/* Z transforms */

void stage_z_normal(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		p[x].z = s[x].z;
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_z_add(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int z = d[x].z;
		z += s[x].z;
		if(z > 65535)
			p[x].z = 65535;
		else
			p[x].z = z;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_z_subtract(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int z = d[x].z;
		z -= s[x].z;
		if(z < 0)
			p[x].z = 0;
		else
			p[x].z = z;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_z_multiply(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		unsigned z = d[x].z;
		z *= s[x].z;
		p[x].z = z >> 12;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}


/* Z-buffering modes */

void stage_zb_above(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x, sx;
	x = sx = 0;
	while(x < w)
	{
		/* Find start of visible span */
		while((x < w) && (p[x].z < d[x].z))
			++x;
		sx = x;

		/* Find end of visible span */
		while((x < w) && (p[x].z >= d[x].z))
			++x;

		/* Render span */
		if(x != sx)
		{
			pipe->pipebuf = p + sx;
			zs_PipeRun(pipe, current + 1, s + sx, d + sx, x - sx);
		}
	}
	pipe->pipebuf = p;
}

void stage_zb_below(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x, sx;
	x = sx = 0;
	while(x < w)
	{
		/* Find start of visible span */
		while((x < w) && (p[x].z > d[x].z))
			++x;
		sx = x;

		/* Find end of visible span */
		while((x < w) && (p[x].z <= d[x].z))
			++x;

		/* Render span */
		if(x != sx)
		{
			pipe->pipebuf = p + sx;
			zs_PipeRun(pipe, current + 1, s + sx, d + sx, x - sx);
		}
	}
	pipe->pipebuf = p;
}


/* Color transforms */

void stage_c_normal(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		p[x].r = s[x].r;
		p[x].g = s[x].g;
		p[x].b = s[x].b;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}


/* Alpha transforms */

void stage_a_fixed(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		p[x].a = pipe->alpha;
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_a_source(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		p[x].a = s[x].a * pipe->alpha >> 8;
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_a_dest(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		p[x].a = d[x].a * pipe->alpha >> 8;
	zs_PipeRun(pipe, current + 1, s, d, w);
}

	
/* Intensity transforms */

void stage_i_fixed(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		p[x].i = pipe->intensity;
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_i_source(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		p[x].i = s[x].i * pipe->intensity >> 8;
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_i_dest(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		p[x].i = d[x].i * pipe->intensity >> 8;
	zs_PipeRun(pipe, current + 1, s, d, w);
}


/* Color channel write mask */

void stage_cw_(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	if(pipe->mode & ZS_CW_R)
		for(x = 0; x < w; ++x)
		{
			int v = (int)p[x].r * p[x].i * p[x].a >> 8;
			v += (int)d[x].r * (255 - p[x].a);
			if(v > 65535)
				v = 65535;
			p[x].r = v >> 8;
		}
	if(pipe->mode & ZS_CW_G)
		for(x = 0; x < w; ++x)
		{
			int v = (int)p[x].g * p[x].i * p[x].a >> 8;
			v += (int)d[x].g * (255 - p[x].a);
			if(v > 65535)
				v = 65535;
			p[x].g = v >> 8;
		}
	if(pipe->mode & ZS_CW_B)
		for(x = 0; x < w; ++x)
		{
			int v = (int)p[x].b * p[x].i * p[x].a >> 8;
			v += (int)d[x].b * (255 - p[x].a);
			if(v > 65535)
				v = 65535;
			p[x].b = v >> 8;
		}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_cw_rgb(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int v = (int)p[x].r * p[x].i * p[x].a >> 8;
		v += (int)d[x].r * (255 - p[x].a);
		if(v > 65535)
			v = 65535;
		d[x].r = v >> 8;
		v = (int)p[x].g * p[x].i * p[x].a >> 8;
		v += (int)d[x].g * (255 - p[x].a);
		if(v > 65535)
			v = 65535;
		d[x].g = v >> 8;
		v = (int)p[x].b * p[x].i * p[x].a >> 8;
		v += (int)d[x].b * (255 - p[x].a);
		if(v > 65535)
			v = 65535;
		d[x].b = v >> 8;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

	
/* Alpha write modes */
	
void stage_aw_copy(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		d[x].a = p[x].a;
	zs_PipeRun(pipe, current + 1, s, d, w);
}
	
void stage_aw_add(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int a = p[x].a;
		a += d[x].a;
		if(a > 255)
			a = 255;
		d[x].a = a;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_aw_subtract(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int a = d[x].a;
		a -= p[x].a;
		if(a < 0)
			a = 0;
		d[x].a = a;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_aw_multiply(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int a = d[x].a;
		a *= p[x].a;
		d[x].a = a >> 8;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}


/* Intensity write modes */

void stage_iw_copy(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		d[x].i = p[x].i;
	zs_PipeRun(pipe, current + 1, s, d, w);
}
	
void stage_iw_add(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int i = p[x].i;
		i += d[x].i;
		if(i > 65535)
			i = 65535;
		d[x].i = i;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_iw_subtract(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int i = d[x].i;
		i -= p[x].i;
		if(i < 0)
			i = 0;
		d[x].i = i;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_iw_multiply(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int i = d[x].i;
		i *= p[x].i;
		d[x].i = i >> 8;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}

void stage_iw_blend(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
	{
		int i = (int)p[x].i * p[x].a;
		i += (int)d[x].i * (255 - p[x].a);
		d[x].i = i >> 8;
	}
	zs_PipeRun(pipe, current + 1, s, d, w);
}


/* Z write modes */

void stage_zw_off(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	/* NOP */
	/* End of pipe! */
}

void stage_zw_on(ZS_Pipe *pipe, int current, ZS_Pixel *s, ZS_Pixel *d, int w)
{
	ZS_Pixel *p = pipe->pipebuf;
	int x;
	for(x = 0; x < w; ++x)
		d[x].z = p[x].z;
	/* End of pipe! */
}


/* Prepare pipe for work */
int zs_PipePrepare(ZS_Pipe *pipe)
{
	pipe->nstages = 0;
	pipe->pipebuf = pipe->state->pipebuf;

	/* Z transforms */
	switch(pipe->mode & ZS_Z_)
	{
	  case ZS_Z_NORMAL:
		pipe->stages[pipe->nstages++] = stage_z_normal;
		break;
	  case ZS_Z_ADD:
		pipe->stages[pipe->nstages++] = stage_z_add;
		break;
	  case ZS_Z_SUBTRACT:
		pipe->stages[pipe->nstages++] = stage_z_subtract;
		break;
	  case ZS_Z_MULTIPLY:
		pipe->stages[pipe->nstages++] = stage_z_multiply;
		break;
	}

	/* Z-buffering modes */
	switch(pipe->mode & ZS_ZB_)
	{
	  case ZS_ZB_ALL:
		/* Fall through directly to the color transform stage */
		break;
	  case ZS_ZB_ABOVE:
		pipe->stages[pipe->nstages++] = stage_zb_above;
		break;
	  case ZS_ZB_BELOW:
		pipe->stages[pipe->nstages++] = stage_zb_below;
		break;
	}

	/* Color transforms */
	switch(pipe->mode & ZS_C_)
	{
	  case ZS_C_NORMAL:
		pipe->stages[pipe->nstages++] = stage_c_normal;
		break;
#if 0
	ZS_C_ADD =	0x00000010,	/* p.rgb = clamp(d.rgb + s.rgb) */
	ZS_C_SUBTRACT =	0x00000020,	/* p.rgb = clamp(d.rgb - s.rgb) */
	ZS_C_MULTIPLY =	0x00000030,	/* p.rgb = d.rgb * s.rgb */
	ZS_C_DIVIDE =	0x00000040,	/* p.rgb = d.rgb / (1 + s.rgb * 256) */
	ZS_C_LIGHTEN =	0x00000050,	/* p.rgb = max(d.rgb, s.rgb) */
	ZS_C_DARKEN =	0x00000060,	/* p.rgb = min(d.rgb, s.rgb) */
	ZS_C_CONTRAST =	0x00000070,	/* p.rgb = clamp((d.rgb - .5)*(s.rgb * 2) + .5) */
	ZS_C_INVERT =	0x00000080,	/* p.rgb = d.rgb * s.rgb +
							(1 - d.rgb)*(1 - s.rgb) */
	ZS_C_HUE =	0x00000090,	/* p.rgb = hue(d.rgb, get_hue(s.rgb)) */
	ZS_C_SATURATE =	0x000000a0,	/* p.rgb = saturate(d.rgb,
							get_saturation(s.rgb))*/
	ZS_C_COLOR =	0x000000b0,	/* p.rgb = colorize(d.rgb, get_color(s.rgb)) */
	ZS_C_VALUE =	0x000000c0,	/* p.rgb = value(d.rgb, get_value(s.rgb)) */
#endif
	  default:
		return -1;
	}

	/* Alpha transforms */
	switch(pipe->mode & ZS_A_)
	{
	  case ZS_A_FIXED:
		pipe->stages[pipe->nstages++] = stage_a_fixed;
		break;
	  case ZS_A_SOURCE:
		pipe->stages[pipe->nstages++] = stage_a_source;
		break;
	  case ZS_A_DEST:
		pipe->stages[pipe->nstages++] = stage_a_dest;
		break;
	}

	/* Intensity transforms */
	switch(pipe->mode & ZS_I_)
	{
	  case ZS_I_FIXED:
		pipe->stages[pipe->nstages++] = stage_i_fixed;
		break;
	  case ZS_I_SOURCE:
		pipe->stages[pipe->nstages++] = stage_i_source;
		break;
	  case ZS_I_DEST:
		pipe->stages[pipe->nstages++] = stage_i_dest;
		break;
	}

	/* Color channel write mask */
	switch(pipe->mode & ZS_CW_)
	{
	  case 0:
		break;
	  case ZS_CW_RGB:
		pipe->stages[pipe->nstages++] = stage_cw_rgb;
		break;
	  default:
		pipe->stages[pipe->nstages++] = stage_cw_;
		break;
	}

	/* Alpha write modes */
	switch(pipe->mode & ZS_AW_)
	{
	  case ZS_AW_OFF:
		break;
	  case ZS_AW_COPY:
		pipe->stages[pipe->nstages++] = stage_aw_copy;
		break;
	  case ZS_AW_ADD:
		pipe->stages[pipe->nstages++] = stage_aw_add;
		break;
	  case ZS_AW_SUBTRACT:
		pipe->stages[pipe->nstages++] = stage_aw_subtract;
		break;
	  case ZS_AW_MULTIPLY:
		pipe->stages[pipe->nstages++] = stage_aw_multiply;
		break;
	}

	/* Intensity write modes */
	switch(pipe->mode & ZS_IW_)
	{
	  case ZS_IW_OFF:
		break;
	  case ZS_IW_COPY:
		pipe->stages[pipe->nstages++] = stage_iw_copy;
		break;
	  case ZS_IW_ADD:
		pipe->stages[pipe->nstages++] = stage_iw_add;
		break;
	  case ZS_IW_SUBTRACT:
		pipe->stages[pipe->nstages++] = stage_iw_subtract;
		break;
	  case ZS_IW_MULTIPLY:
		pipe->stages[pipe->nstages++] = stage_iw_multiply;
		break;
	  case ZS_IW_BLEND:
		pipe->stages[pipe->nstages++] = stage_iw_blend;
		break;
	}

	/* Z write modes */
	switch(pipe->mode & ZS_ZW_)
	{
	  case ZS_ZW_OFF:
		pipe->stages[pipe->nstages++] = stage_zw_off;
		break;
	  case ZS_ZW_ON:
		pipe->stages[pipe->nstages++] = stage_zw_on;
		break;
	}
	return 0;
}


/*------------------------------------------------------------------
	Pipe API
------------------------------------------------------------------*/

ZS_Pipe *zs_NewPipe(ZS_State *state)
{
	ZS_Pipe *p = (ZS_Pipe *)calloc(1, sizeof(ZS_Pipe));
	if(!p)
		return NULL;
	p->state = state;
	++state->refcount;
	/* Sensible defaults (?) */
	zs_PipeZ(p, ZS_Z_NORMAL | ZS_ZB_ABOVE);
	zs_PipeColor(p, ZS_C_NORMAL);
	zs_PipeAlpha(p, ZS_A_FIXED, 1.0);
	zs_PipeIntensity(p, ZS_I_FIXED, 1.0);
	zs_PipeWrite(p, ZS_CW_RGB | ZS_AW_OFF | ZS_IW_OFF | ZS_ZW_ON);
	return p;
}


int zs_PipeZ(ZS_Pipe *pipe, ZS_pipemode zmode)
{
	pipe->mode &= ~(ZS_Z_ | ZS_ZB_);
	pipe->mode |= zmode & (ZS_Z_ | ZS_ZB_);
	return zs_PipePrepare(pipe);
}


int zs_PipeColor(ZS_Pipe *pipe, ZS_pipemode cmode)
{
	pipe->mode &= ~ZS_C_;
	pipe->mode |= cmode & ZS_C_;
	return zs_PipePrepare(pipe);
}


int zs_PipeAlpha(ZS_Pipe *pipe, ZS_pipemode amode, float alpha)
{
	pipe->mode &= ~ZS_A_;
	pipe->mode |= amode & ZS_A_;
	pipe->alpha = alpha * 255.0;
	return zs_PipePrepare(pipe);
}


int zs_PipeIntensity(ZS_Pipe *pipe, ZS_pipemode imode, float intensity)
{
	pipe->mode &= ~ZS_I_;
	pipe->mode |= imode & ZS_I_;
	pipe->intensity = intensity * 256.0;
	return zs_PipePrepare(pipe);
}


int zs_PipeWrite(ZS_Pipe *pipe, ZS_pipemode wmode)
{
	pipe->mode &= ~(ZS_CW_ |ZS_AW_ |ZS_IW_ | ZS_ZW_);
	pipe->mode |= wmode & (ZS_CW_ |ZS_AW_ |ZS_IW_ | ZS_ZW_);
	return zs_PipePrepare(pipe);
}


void zs_FreePipe(ZS_Pipe *pipe)
{
	zs_Close(pipe->state);
	free(pipe);
}
